﻿using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.StaticFiles;
using SimpleFileServer.Dtos;
using System.Text.RegularExpressions;

namespace SimpleFileServer.Controllers
{
    /// <summary>
    /// Controller for a very rudimentary API for file storage
    /// </summary>
    [ApiController]
    [Authorize]
    public class FileServerController : ControllerBase
    {
        private const string ContentDirectory = "HttpRoot";
        private readonly string ServerPath = Path.Combine(Directory.GetCurrentDirectory(), ContentDirectory);

        private IWebHostEnvironment Environment { get; }

        /// <summary>
        /// Creates an instance of this controller
        /// </summary>
        /// <param name="environment">WebHostEnvironment instance running this controller</param>
        public FileServerController(IWebHostEnvironment environment)
        {
            Environment = environment;

            if (!Directory.Exists(ServerPath))
                Directory.CreateDirectory(ServerPath);
        }

        private string GetPath(string path, bool absolute = true) => (absolute)
            ? (path == "/" ? ServerPath : Path.Combine(ServerPath, path))
            : Path.Combine(ContentDirectory, path);

        private static string GetMimeType(string path)
        {
            new FileExtensionContentTypeProvider().TryGetContentType(path, out string? contentType);
            return contentType ?? "application/octet-stream";
        }

        private static bool CheckPathSanity(string path) => !Regex.Match(path, @"/?\.\./?").Success;

        /// <summary>
        /// Downloads the file specified by the path paramenter
        /// </summary>
        /// <param name="path">the path of the file</param>
        /// <returns></returns>
        [HttpGet]
        [Route("files/{**path}")]
        [Authorize(Roles = "download")]
        public IActionResult Get([FromRoute] string? path)
        {
            if (path is null || !CheckPathSanity(path))
                return BadRequest("Invalid path");

            string filePath = GetPath(path);

            if (!System.IO.File.Exists(filePath))
            {
                return NotFound();
            }

            var stream = Environment
                .ContentRootFileProvider
                .GetFileInfo(GetPath(path, absolute: false))
                .CreateReadStream();

            return File(stream, GetMimeType(path), true);
        }

        /// <summary>
        /// Downloads the file specified by the path paramenter
        /// </summary>
        /// <param name="path">the path of the file</param>
        /// <returns></returns>
        [HttpHead]
        [Route("files/{**path}")]
        [Authorize(Roles = "download")]
        public IActionResult GetHead([FromRoute] string? path)
        {
            if (path is null || !CheckPathSanity(path))
                return BadRequest("Invalid path");

            string filePath = GetPath(path);

            if (!System.IO.File.Exists(filePath))
            {
                return NotFound();
            }

            var stream = Environment
                .ContentRootFileProvider
                .GetFileInfo(GetPath(path, absolute: false))
                .CreateReadStream();

            Response.Headers.Add("Content-Length", stream.Length.ToString());
            Response.Headers.Add("Content-Type", "application/octet-stream");

            return Ok();
        }

        [HttpGet]
        [Route("browse/{**path=/}")]
        [Authorize(Roles = "list")]
        public IActionResult ListDirectory([FromRoute] string path)
        {
            string targetPath = GetPath(path);
            if (!Directory.Exists(targetPath))
            {
                return NotFound();
            }

            var contents = Directory.GetFiles(targetPath, "", SearchOption.AllDirectories);

            var fileInfo = new List<FileInfoDto>();



            foreach (var file in contents)
            {
                fileInfo.Add(new FileInfoDto
                {
                    Name = Path.GetFileName(file),
                    Path = Path.GetRelativePath(ServerPath, Path.GetDirectoryName(file)) is var _path && _path == "." ? string.Empty : _path,
                    Size = (new FileInfo(file)).Length,
                    MimeType = GetMimeType(file),
                });
            }

            return Ok(fileInfo);
        }

        [HttpPost]
        [Route("files/{**path}")]
        [Authorize(Roles = "upload")]
        [RequestSizeLimit(1073741824)]
        public async Task<IActionResult> Upload([FromRoute] string path, [FromBody] Stream data)
        {
            if (data.Length == 0)
                return BadRequest("File is empty");

            if (!CheckPathSanity(path))
                return BadRequest("Invalid path");

            string proposedPath = GetPath(path);
            string? proposedPathDirectory = Path.GetDirectoryName(proposedPath)
                ?? throw new InvalidOperationException("Proposed path is null");
            if (!Directory.Exists(proposedPathDirectory))
            {
                Directory.CreateDirectory(proposedPathDirectory);
            }

            using var stream = System.IO.File.OpenWrite(proposedPath);
            data.Seek(0, SeekOrigin.Begin);

            await data.CopyToAsync(stream);

            return CreatedAtAction(nameof(Get), new { path = path }, string.Empty);
        }
    }
}
